###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

# Import Pyomo libraries
from pyomo.environ import (Block,
                           Var,
                           Suffix,
                           units as pyunits,
                           ExternalFunction)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData)
from idaes.core.util import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.functions import functions_lib
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

# From watertap
from watertap.unit_models.mvc.components.complete_condenser import Condenser


_log = idaeslog.getLogger(__name__)

@declare_process_block_class("Evaporator")
class EvaporatorData(UnitModelBlockData):
    """
    Evaporator model for MVC
    """
    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False."""))
    CONFIG.declare("property_package_feed", ConfigValue(
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args_feed", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}"""))
    CONFIG.declare("property_package_vapor", ConfigValue(
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args_vapor", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}"""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))


    def build(self):
        super().build()

        if self.config.property_package_feed is None:
            raise ConfigurationError("Users must provide a feed property package to the evaporator unit model")
        if self.config.property_package_vapor is None:
            raise ConfigurationError("Users must provide a vapor property package to the evaporator unit model")

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta_feed = self.config.property_package_feed.get_metadata().get_derived_units

        # Add shared unit model variables
        self.U = Var(
            initialize=1e3,
            bounds=(10, 1e4),
            units=pyunits.J * pyunits.s ** -1 * pyunits.m ** -2 * pyunits.K ** -1
        )

        self.area = Var(
            initialize=1e2,
            bounds=(1e-1, 1e4),
            units=pyunits.m ** 2
        )

        self.delta_temperature_in = Var(
            initialize=1e1,
            bounds=(1e-8, 1e3),
            units=pyunits.K
        )

        self.delta_temperature_out = Var(
            initialize=1e1,
            bounds=(1e-8, 1e3),
            units=pyunits.K
        )

        self.lmtd = Var(
            initialize=1e1,
            bounds=(1e-8, 1e3),
            units=pyunits.K
        )

        # Add feed_side block
        self.feed_side = Block()

        # Add unit variables to feed
        self.feed_side.heat_transfer = Var(
            initialize = 1e4,
            bounds=(1,1e10),
            units=pyunits.J * pyunits.s**-1
        )

        # Add feed_side state blocks
        # Feed state block
        tmp_dict = dict(**self.config.property_package_args_feed)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package_feed
        tmp_dict["defined_state"] = True  # feed inlet defined
        self.feed_side.properties_feed = self.config.property_package_feed.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of feed inlet",
            default=tmp_dict)

        # Brine state block
        tmp_dict["defined_state"] = False  # brine outlet not yet defined
        self.feed_side.properties_brine = self.config.property_package_feed.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of brine outlet",
            default=tmp_dict)

        # Vapor state block
        tmp_dict = dict(**self.config.property_package_args_vapor)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package_vapor
        tmp_dict["defined_state"] = False  # vapor outlet not yet defined
        self.feed_side.properties_vapor = self.config.property_package_vapor.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of vapor outlet",
            default=tmp_dict)

        # Add condenser
        self.condenser = Condenser(default={"property_package": self.config.property_package_vapor})

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name='inlet_feed', block=self.feed_side.properties_feed)
        self.add_port(name='outlet_brine', block=self.feed_side.properties_brine)
        self.add_port(name='outlet_vapor', block=self.feed_side.properties_vapor)
        self.add_port(name='inlet_condenser', block=self.condenser.control_volume.properties_in)
        self.add_port(name='outlet_condenser', block=self.condenser.control_volume.properties_out)


        ### FEED SIDE CONSTRAINTS ###
        # Mass balance
        @self.feed_side.Constraint(self.flowsheet().time, self.config.property_package_feed.component_list, doc="Mass balance")
        def eq_mass_balance(b,t,j):
            lb = b.properties_vapor[t].flow_mass_phase_comp['Liq', 'H2O'].lb
            b.properties_vapor[t].flow_mass_phase_comp['Liq', 'H2O'].fix(lb)
            if j == 'H2O':
                return b.properties_feed[t].flow_mass_phase_comp['Liq','H2O']  == \
                   b.properties_brine[t].flow_mass_phase_comp['Liq','H2O'] + b.properties_vapor[t].flow_mass_phase_comp['Vap','H2O']
            else:
                return b.properties_feed[t].flow_mass_phase_comp['Liq',j] == b.properties_brine[t].flow_mass_phase_comp['Liq',j]

        # Energy balance
        @self.feed_side.Constraint(self.flowsheet().time,doc="Energy balance")
        def eq_energy_balance(b,t):
            return b.heat_transfer + b.properties_feed[t].enth_flow == \
                   b.properties_brine[t].enth_flow + b.properties_vapor[t].enth_flow_phase['Vap']

        # Brine pressure
        @self.feed_side.Constraint(self.flowsheet().time, doc="Brine pressure")
        def eq_brine_pressure(b, t):
             return b.properties_brine[t].pressure == b.properties_brine[t].pressure_sat

        # Vapor pressure
        @self.feed_side.Constraint(self.flowsheet().time,doc="Vapor pressure")
        def eq_vapor_pressure(b,t):
            return b.properties_vapor[t].pressure == b.properties_brine[t].pressure

        # Vapor temperature
        @self.feed_side.Constraint(self.flowsheet().time, doc="Vapor temperature")
        def eq_vapor_temperature(b, t):
            return b.properties_vapor[t].temperature == b.properties_brine[t].temperature
            # return b.properties_vapor[t].temperature == 0.5*(b.properties_out[t].temperature + b.properties_in[t].temperature)

        ### EVAPORATOR CONSTRAINTS ###
        # Temperature difference in
        @self.Constraint(self.flowsheet().time, doc="Temperature difference in")
        def eq_delta_temperature_in(b, t):
            return b.delta_temperature_in == b.condenser.control_volume.properties_in[t].temperature - b.feed_side.properties_brine[t].temperature

        # Temperature difference out
        @self.Constraint(self.flowsheet().time, doc="Temperature difference out")
        def eq_delta_temperature_out(b, t):
            return b.delta_temperature_out == b.condenser.control_volume.properties_out[t].temperature - b.feed_side.properties_brine[t].temperature

        #log mean temperature
        @self.Constraint(self.flowsheet().time, doc="Log mean temperature difference")
        def eq_lmtd(b, t):
            dT_in= b.delta_temperature_in
            dT_out = b.delta_temperature_out
            temp_units = pyunits.get_units(dT_in)
            dT_avg = (dT_in + dT_out) / 2
            # external function that ruturns the real root, for the cuberoot of negitive
            # numbers, so it will return without error for positive and negitive dT.
            b.cbrt = ExternalFunction(
                library=functions_lib(),
                function="cbrt",
                arg_units=[temp_units ** 3])
            return b.lmtd == b.cbrt((dT_in * dT_out * dT_avg))* temp_units

        #Heat transfer between feed side and condenser
        @self.Constraint(self.flowsheet().time,doc="Heat transfer balance")
        def eq_heat_balance(b,t):
            return b.feed_side.heat_transfer == -b.condenser.control_volume.heat[t]

        #Evaporator heat transfer
        @self.Constraint(self.flowsheet().time,doc="Evaporator heat transfer")
        def eq_evaporator_heat(b,t):
            return b.feed_side.heat_transfer == b.U*b.area*b.lmtd

    def initialize(
            blk,
            state_args=None,
            outlvl=idaeslog.NOTSET,
            solver=None,
            optarg=None):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize feed side
        flags_feed = blk.feed_side.properties_feed.initialize(solver=solver, optarg=optarg, hold_state=True)
        init_log.info_high("Initialization Step 1 Complete.")
        # # ---------------------------------------------------------------------
        # # Initialize brine
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = blk.feed_side.properties_feed[
                blk.flowsheet().config.time.first()].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        blk.feed_side.properties_brine.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args
        )

        state_args_vapor = {}
        state_args_vapor['pressure'] = 0.5*state_args['pressure']
        state_args_vapor['temperature'] = state_args['temperature']
        state_args_vapor['flow_mass_phase_comp'] = {
            ('Liq','H2O'): blk.feed_side.properties_vapor[0].flow_mass_phase_comp['Liq','H2O'].lb,
            ('Vap','H2O'): state_args['flow_mass_phase_comp'][('Liq','H2O')]}

        blk.feed_side.properties_vapor.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_vapor,
        )

        init_log.info_high("Initialization Step 2 Complete.")

        # intialize condenser
        state_args_condenser=state_args_vapor
        state_args_condenser['flow_mass_phase_comp'][('Vap', 'H2O')] \
            = 0.5*state_args_condenser['flow_mass_phase_comp'][('Vap', 'H2O')]
        state_args_condenser['pressure'] = blk.feed_side.properties_brine[0].pressure_sat.value
        state_args_condenser['temperature'] = state_args['temperature'] + 5
        blk.condenser.initialize(state_args=state_args_condenser)
        # assert False
        # flags_condenser_cv = blk.condenser.initialize(state_args=state_args_condenser,hold_state=True)
        init_log.info_high("Initialization Step 3 Complete.")
        # ---------------------------------------------------------------------
        # Deactivate heat transfer balance
        # blk.eq_heat_balance.deactivate()

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
            "Initialization Step 4 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release feed and condenser inlet states
        blk.feed_side.properties_feed.release_state(flags_feed, outlvl=outlvl)
        #blk.condenser.control_volume.release_state(flags_condenser_cv, outlvl=outlvl)

        init_log.info(
            "Initialization Complete: {}".format(idaeslog.condition(res))
        )

    def _get_performance_contents(self, time_point=0):
        var_dict = {"Heat transfer": self.feed_side.heat_transfer,
                    "Evaporator temperature": self.feed_side.properties_brine[0].temperature,
                    "Evaporator pressure": self.feed_side.properties_brine[0].pressure}

        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        self.condenser.calculate_scaling_factors()

        if iscale.get_scaling_factor(self.feed_side.heat_transfer) is None:
            sf = iscale.get_scaling_factor(self.feed_side.properties_vapor[0].enth_flow_phase['Vap'])
            iscale.set_scaling_factor(self.feed_side.heat_transfer, sf)

        # for (t,j), c in self.eq_mass_balance.items():
        #     sf = iscale.get_scaling_factor(self.properties_feed[t].flow_mass_phase_comp['Liq', j])
        #     iscale.constraint_scaling_transform(c, sf)
        #
        # # Pressure constraints
        # sf = iscale.get_scaling_factor(self.properties_feed[0].pressure)
        # iscale.constraint_scaling_transform(self.eq_vapor_pressure, sf)
        #
        # # Temperature constraint
        # sf = iscale.get_scaling_factor(self.properties_feed[0].temperature)
        # iscale.constraint_scaling_transform(self.eq_brine_temperature, sf)
        # iscale.constraint_scaling_transform(self.eq_vapor_temperature, sf)
        #
        # # Efficiency, work constraints
        # sf = iscale.get_scaling_factor(self.heat_transfer)
        # iscale.constraint_scaling_transform(self.eq_energy_balance, sf)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains common methods for determining the electricity intensity
and demand for zero-order unit models.
"""

import idaes.logger as idaeslog

from pyomo.environ import Param, Var, units as pyunits

# Some more inforation about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


def _common(self):
    # Add electricity consumption to model
    self.electricity = Var(self.flowsheet().time,
                           units=pyunits.kW,
                           bounds=(0, None),
                           doc="Electricity consumption of unit")

    self._perf_var_dict["Electricity Demand"] = self.electricity


def constant_intensity(self):
    """
    Helper method for implementing electricity demand assuming constant
    intensity based on the inlet volumetric flow rate.

    E[t] = Q[t] * intensity

    Two variables are added to the model:
        * electricity (indexed by time)
        * energy_electric_flow_vol_inlet (unindexed)

    One constraint is added to the model:
        * electricity_consumption (indexed by time)
    """
    _common(self)

    self.energy_electric_flow_vol_inlet = Var(
        units=pyunits.kWh/pyunits.m**3,
        doc="Electricity intensity with respect to inlet flowrate of unit")

    @self.Constraint(self.flowsheet().time,
                     doc='Constraint for electricity consumption based on '
                     'feed flowrate.')
    def electricity_consumption(b, t):
        return b.electricity[t] == (
            b.energy_electric_flow_vol_inlet *
            pyunits.convert(b.get_inlet_flow(t),
                            to_units=pyunits.m**3/pyunits.hour))

    self._fixed_perf_vars.append(self.energy_electric_flow_vol_inlet)
    self._perf_var_dict["Electricity Intensity"] = \
        self.energy_electric_flow_vol_inlet


def pump_electricity(self, flow_rate):
    """
    Helper method for calculating electricity demand based on a
    pump flow equation.

    E[t] = (0.746 * Q[t] * H / (3960 * eta_pump * eta_motor))

    Here Q is the volumetric flowrate to be used to calculate electricity demand,
    H is the lift height for the pump and eta_pump and eta_motor are the
    efficiencies of the pump and motor respectively.

    Args:
        flow_rate - term to use for Q in the equation above.
    """
    _common(self)

    self.lift_height = Param(initialize=100,
                             units=pyunits.feet,
                             mutable=True,
                             doc="Lift height for pump")
    self.eta_pump = Param(initialize=0.9,
                          units=pyunits.dimensionless,
                          mutable=True,
                          doc="Efficiency of pump")
    self.eta_motor = Param(initialize=0.9,
                           units=pyunits.dimensionless,
                           mutable=True,
                           doc="Efficiency of motor")

    @self.Constraint(self.flowsheet().time,
                     doc='Constraint for electricity consumption based on '
                     'pump flowrate.')
    def electricity_consumption(b, t):
        A = 3960*pyunits.gallon*pyunits.foot/pyunits.minute/pyunits.horsepower
        return b.electricity[t] == pyunits.convert(
            flow_rate[t]*self.lift_height/(A*self.eta_pump*self.eta_motor),
            to_units=pyunits.kW)



###############################################################################
# ProteusLib Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/nawi-hub/proteuslib/"
#
###############################################################################
import pytest
from watertap.property_models.coagulation_prop_pack import CoagulationParameterBlock
from watertap.property_models.NaCl_prop_pack import NaClParameterBlock
from watertap.property_models.seawater_prop_pack import SeawaterParameterBlock
from watertap.unit_models.coag_floc_model import CoagulationFlocculation
from pyomo.environ import (ConcreteModel,
                           assert_optimal_termination,
                           value,
                           Set,
                           Param,
                           Var,
                           units as pyunits,
                           Suffix,
                           Constraint,
                           SolverFactory,
                           SolverStatus,
                           TerminationCondition)
from idaes.core import (FlowsheetBlock,
                        MaterialFlowBasis,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
import idaes.core.util.scaling as iscale
from idaes.core.util.scaling import badly_scaled_var_generator
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver
import re

__author__ = "Austin Ladshaw"

solver = get_solver()

# -----------------------------------------------------------------------------
# Start test class
class TestCoagulation_withChemicals():
    @pytest.fixture(scope="class")
    def coag_obj_w_chems(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.properties = CoagulationParameterBlock()
        ## NOTE: These values provided are just DUMMY values for the purposes
        #        of testing. They are not meant to be representative of any
        #        particular chemicals or real-world additives.
        chem_dict = {'Alum':
                        {"parameter_data":
                            {"mw_additive": (200, pyunits.g/pyunits.mol),
                            "moles_salt_per_mole_additive": 3,
                            "mw_salt": (100, pyunits.g/pyunits.mol)}
                        },
                      'Poly':
                        {"parameter_data":
                            {"mw_additive": (25, pyunits.g/pyunits.mol),
                            "moles_salt_per_mole_additive": 0,
                            "mw_salt": (23, pyunits.g/pyunits.mol)}
                        }
                     }
        model.fs.unit = CoagulationFlocculation(default={
            "property_package": model.fs.properties,
            "chemical_additives": chem_dict })

        return model

    @pytest.mark.unit
    def test_build_model(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        assert len(model.fs.unit.config.chemical_additives) == 2
        assert isinstance(model.fs.unit.slope, Var)
        assert isinstance(model.fs.unit.intercept, Var)
        assert isinstance(model.fs.unit.initial_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.final_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.chemical_doses, Var)
        assert len(model.fs.unit.chemical_doses) == 2
        assert isinstance(model.fs.unit.chemical_mw, Param)
        assert len(model.fs.unit.chemical_mw) == 2
        assert isinstance(model.fs.unit.salt_mw, Param)
        assert len(model.fs.unit.salt_mw) == 2
        assert isinstance(model.fs.unit.salt_from_additive_mole_ratio, Param)
        assert len(model.fs.unit.salt_from_additive_mole_ratio) == 2
        assert isinstance(model.fs.unit.tss_loss_rate, Var)
        assert isinstance(model.fs.unit.eq_tss_loss_rate, Constraint)

        assert isinstance(model.fs.unit.tds_gain_rate, Var)
        assert isinstance(model.fs.unit.eq_tds_gain_rate, Constraint)

        assert isinstance(model.fs.unit.eq_mass_transfer_term, Constraint)

    @pytest.mark.unit
    def test_stats(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(model)
        assert degrees_of_freedom(model) == 11

        # set the operational parameters
        model.fs.unit.fix_tss_turbidity_relation_defaults()
        model.fs.unit.initial_turbidity_ntu.fix()
        model.fs.unit.final_turbidity_ntu.fix(5)
        model.fs.unit.chemical_doses[0, 'Alum'].fix(10)
        model.fs.unit.chemical_doses[0, 'Poly'].fix(5)

        # set the inlet streams
        assert degrees_of_freedom(model) == 6
        model.fs.unit.inlet.pressure.fix(101325)
        model.fs.unit.inlet.temperature.fix(298.15)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(1)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'TDS'].fix(0.01)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'TSS'].fix(0.01)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'Sludge'].fix(0.0)

        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_scaling(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        # Set some scaling factors and look for 'bad' scaling
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'TSS'))
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'TDS'))
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Sludge'))
        iscale.calculate_scaling_factors(model.fs)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(model))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
        var.name: val
        for (var, val) in iscale.badly_scaled_var_generator(
            model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, coag_obj_w_chems):
        model = coag_obj_w_chems
        initialization_tester(model)

        # check to make sure DOF does not change
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
        var.name: val
        for (var, val) in iscale.badly_scaled_var_generator(
            model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'H2O']) == pytest.approx(1,  rel=1e-4)
        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'Sludge']) == pytest.approx(0.00999,  rel=1e-4)
        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'TDS']) == pytest.approx(0.010015,  rel=1e-4)
        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'TSS']) == pytest.approx(9.36352e-06,  rel=1e-4)


# -----------------------------------------------------------------------------
# Start test class without chemicals added
class TestCoagulation_withNoChemicals():
    @pytest.fixture(scope="class")
    def coag_obj_wo_chems(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.properties = CoagulationParameterBlock()
        model.fs.unit = CoagulationFlocculation(default={
            "property_package": model.fs.properties})

        return model

    @pytest.mark.unit
    def test_build_model(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        assert len(model.fs.unit.config.chemical_additives) == 0
        assert isinstance(model.fs.unit.slope, Var)
        assert isinstance(model.fs.unit.intercept, Var)
        assert isinstance(model.fs.unit.initial_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.final_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.chemical_doses, Var)
        assert len(model.fs.unit.chemical_doses) == 0
        assert isinstance(model.fs.unit.chemical_mw, Param)
        assert len(model.fs.unit.chemical_mw) == 0
        assert isinstance(model.fs.unit.salt_mw, Param)
        assert len(model.fs.unit.salt_mw) == 0
        assert isinstance(model.fs.unit.salt_from_additive_mole_ratio, Param)
        assert len(model.fs.unit.salt_from_additive_mole_ratio) == 0
        assert isinstance(model.fs.unit.tss_loss_rate, Var)
        assert isinstance(model.fs.unit.eq_tss_loss_rate, Constraint)

        assert not hasattr(model.fs.unit, "tds_gain_rate")
        assert not hasattr(model.fs.unit, "eq_tds_gain_rate")

        assert isinstance(model.fs.unit.eq_mass_transfer_term, Constraint)

    @pytest.mark.unit
    def test_stats(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(model)
        assert degrees_of_freedom(model) == 9

        # set the operational parameters
        model.fs.unit.fix_tss_turbidity_relation_defaults()
        model.fs.unit.initial_turbidity_ntu.fix(5000)
        model.fs.unit.final_turbidity_ntu.fix(100)

        # set the inlet streams
        assert degrees_of_freedom(model) == 6

        tss_in = value(model.fs.unit.compute_inlet_tss_mass_flow(0))
        assert tss_in == pytest.approx(0.0093,  rel=1e-4)
        model.fs.unit.inlet.pressure.fix(101325)
        model.fs.unit.inlet.temperature.fix(298.15)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(1)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'TDS'].fix(0.01)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'TSS'].fix(tss_in)
        model.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'Sludge'].fix(0.0)

        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_scaling(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        # Set some scaling factors and look for 'bad' scaling
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'TSS'))
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'TDS'))
        model.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Sludge'))
        #iscale.set_scaling_factor(model.fs.unit.tss_loss_rate, 100)
        iscale.calculate_scaling_factors(model.fs)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(model))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
        var.name: val
        for (var, val) in iscale.badly_scaled_var_generator(
            model, large=1e2, small=1e-2
            )
        }
        print(iscale.get_scaling_factor(model.fs.unit.tss_loss_rate))
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems
        initialization_tester(model)

        # check to make sure DOF does not change
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
        var.name: val
        for (var, val) in iscale.badly_scaled_var_generator(
            model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'H2O']) == pytest.approx(1,  rel=1e-4)
        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'Sludge']) == pytest.approx(0.0091127,  rel=1e-4)
        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'TDS']) == pytest.approx(0.01,  rel=1e-4)
        assert value(model.fs.unit.outlet.flow_mass_phase_comp[0, 'Liq', 'TSS']) == pytest.approx(0.00018725,  rel=1e-4)

# -----------------------------------------------------------------------------
# Start test class with bad config
class TestCoagulation_withBadConfig():
    @pytest.fixture(scope="class")
    def coag_obj_bad_config(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.properties = CoagulationParameterBlock()

        return model

    @pytest.mark.unit
    def test_build_model_catch_errors(self, coag_obj_bad_config):
        model = coag_obj_bad_config

        bad_dict1 = {'Alum':
                        {"foo_bar":
                            {"mw_additive": (200, pyunits.g/pyunits.mol),
                            "moles_salt_per_mole_additive": 3,
                            "mw_salt": (100, pyunits.g/pyunits.mol)}
                        }
                     }
        with pytest.raises(ConfigurationError, match="Did not provide a 'parameter_data' for chemical"):
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties,
                "chemical_additives": bad_dict1 })

        bad_dict2 = {'Alum':
                        {"parameter_data":
                            {"foo_bar": (200, pyunits.g/pyunits.mol),
                            "moles_salt_per_mole_additive": 3,
                            "mw_salt": (100, pyunits.g/pyunits.mol)}
                        }
                     }
        with pytest.raises(ConfigurationError, match="Did not provide a 'mw_additive' for chemical"):
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties,
                "chemical_additives": bad_dict2 })

        bad_dict3 = {'Alum':
                        {"parameter_data":
                            {"mw_additive": (200, pyunits.g/pyunits.mol),
                            "moles_salt_per_mole_additive": "foo-bar",
                            "mw_salt": (100, pyunits.g/pyunits.mol)}
                        }
                     }
        with pytest.raises(ConfigurationError, match="Did not provide a number for 'moles_salt_per_mole_additive'"):
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties,
                "chemical_additives": bad_dict3 })

        bad_dict4 = {'Alum':
                        {"parameter_data":
                            {"mw_additive": (200, pyunits.g/pyunits.mol),
                            "foo-bar": 3,
                            "mw_salt": (100, pyunits.g/pyunits.mol)}
                        }
                     }
        with pytest.raises(ConfigurationError, match="Did not provide a 'moles_salt_per_mole_additive' for chemical"):
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties,
                "chemical_additives": bad_dict4 })

        bad_dict5 = {'Alum':
                        {"parameter_data":
                            {"mw_additive": (200, pyunits.g/pyunits.mol),
                            "moles_salt_per_mole_additive": 3,
                            "foo-bar": (100, pyunits.g/pyunits.mol)}
                        }
                     }
        with pytest.raises(ConfigurationError, match="Did not provide a 'mw_salt' for chemical"):
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties,
                "chemical_additives": bad_dict5 })

        bad_dict6 = {'Alum':
                        {"parameter_data":
                            {"mw_additive": "not a tuple",
                            "moles_salt_per_mole_additive": 3,
                            "mw_salt": (100, pyunits.g/pyunits.mol)}
                        }
                     }
        with pytest.raises(ConfigurationError, match="Did not provide a tuple for 'mw_additive'"):
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties,
                "chemical_additives": bad_dict6 })

# -----------------------------------------------------------------------------
# Start test class with bad config
class TestCoagulation_withBadProperties():
    @pytest.fixture(scope="class")
    def coag_obj_bad_properties(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})

        return model

    @pytest.mark.unit
    def test_build_model_catch_prop_errors(self, coag_obj_bad_properties):
        model = coag_obj_bad_properties

        error_msg = ("Coagulation-Flocculation model MUST contain ('Liq','TDS') "
                    "as a component, but the property package has only specified "
                    "the following components [('Liq', 'H2O'), ('Liq', 'NaCl')]")
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            model.fs.properties = NaClParameterBlock()
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties })

        error_msg = ("Coagulation-Flocculation model MUST contain ('Liq','Sludge') "
                    "as a component, but the property package has only specified "
                    "the following components [('Liq', 'H2O'), ('Liq', 'TDS')]")
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            model.fs.properties = SeawaterParameterBlock()
            model.fs.unit = CoagulationFlocculation(default={
                "property_package": model.fs.properties })

        # NOTE: package must also contain ('Liq','TSS') as a component



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order sedimentation model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    Block, ConcreteModel, Constraint, value, Var, assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import SedimentationZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestSedimentationZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tss", "foo"]})

        m.fs.unit = SedimentationZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

        assert isinstance(model.fs.unit.settling_velocity, Var)
        assert isinstance(model.fs.unit.basin_surface_area, Var)
        assert isinstance(model.fs.unit.basin_surface_area_constraint,
                          Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("sedimentation")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.settling_velocity[0].fixed
        assert model.fs.unit.settling_velocity[0].value == data[
            "settling_velocity"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.4e-2, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(214.2857, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(71.4286, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["foo"]))
        assert (pytest.approx(1.10265e-2, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(2.4960, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(90.690, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(2.9735e-3, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(999.66, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0, abs=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (pytest.approx(30.139, rel=1e-5) ==
                value(model.fs.unit.basin_surface_area[0]))


    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                       : Value      : Fixed : Bounds
    Basin Surface Area (ft^2) :     30.139 : False : (None, None)
           Electricity Demand : 8.0000e-10 : False : (0, None)
        Electricity Intensity :     0.0000 :  True : (None, None)
      Settling Velocity (m/s) :  0.0050000 :  True : (None, None)
         Solute Removal [foo] :     0.0000 :  True : (0, None)
         Solute Removal [tss] :    0.99083 :  True : (0, None)
               Water Recovery :    0.99990 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet   Treated  Byproduct
    Volumetric Flowrate    0.014000 0.011027  0.0029735
    Mass Concentration H2O   714.29   906.81    0.33631
    Mass Concentration tss   214.29   2.4960     999.66
    Mass Concentration foo   71.429   90.690 2.6905e-07
====================================================================================
"""

        assert output in stream.getvalue()


class TestSedimentationZO_phosphorus_capture_tss:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tss"]})

        m.fs.unit = SedimentationZO(default={
            "property_package": m.fs.params,
            "database": m.db,
            "process_subtype": "phosphorus_capture"})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

        assert isinstance(model.fs.unit.settling_velocity, Var)
        assert isinstance(model.fs.unit.basin_surface_area, Var)
        assert isinstance(model.fs.unit.basin_surface_area_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.final_phosphate_mass, Var)
        assert isinstance(model.fs.unit.phosphate_mass_flow_constraint, Constraint)
        assert isinstance(model.fs.unit.phosphorus_solids_ratio, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("sedimentation",
                                                      subtype="phosphorus_capture")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.settling_velocity[0].fixed
        assert model.fs.unit.settling_velocity[0].value == data[
            "settling_velocity"]["value"]

        assert model.fs.unit.phosphorus_solids_ratio[0].fixed
        assert model.fs.unit.phosphorus_solids_ratio[0].value == data[
            "phosphorus_solids_ratio"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.009864, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.3041363, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.0031360, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(955.68, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (pytest.approx(27.986, rel=1e-5) ==
                value(model.fs.unit.basin_surface_area[0]))
        assert (pytest.approx(0.44955, rel=1e-5) ==
                value(model.fs.unit.final_phosphate_mass[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                         : Value      : Fixed : Bounds
                      Basin Surface Area (ft^2) :     27.986 : False : (None, None)
                             Electricity Demand : 7.0000e-10 : False : (0, None)
                          Electricity Intensity :     0.0000 :  True : (None, None)
    Final mass flow of settled phosphate (kg/s) :    0.44955 : False : (None, None)
                Phosphorus-Solids Ratio (kg/kg) :    0.15000 :  True : (None, None)
                        Settling Velocity (m/s) :  0.0050000 :  True : (None, None)
                           Solute Removal [tss] :    0.99900 :  True : (0, None)
                                 Water Recovery :    0.98610 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet    Treated  Byproduct
    Volumetric Flowrate    0.013000 0.0098640 0.0031360 
    Mass Concentration H2O   769.23    999.70    44.324 
    Mass Concentration tss   230.77   0.30414    955.68 
====================================================================================
"""

        assert output in stream.getvalue()


class TestSedimentationZO_phosphorus_capture_phosphates:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["phosphates"]})

        m.fs.unit = SedimentationZO(default={
            "property_package": m.fs.params,
            "database": m.db,
            "process_subtype": "phosphorus_capture"})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "phosphates"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

        assert isinstance(model.fs.unit.settling_velocity, Var)
        assert isinstance(model.fs.unit.basin_surface_area, Var)
        assert isinstance(model.fs.unit.basin_surface_area_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.final_solids_mass, Var)
        assert isinstance(model.fs.unit.solids_mass_flow_constraint, Constraint)
        assert isinstance(model.fs.unit.phosphorus_solids_ratio, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("sedimentation",
                                                      subtype="phosphorus_capture")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.settling_velocity[0].fixed
        assert model.fs.unit.settling_velocity[0].value == data[
            "settling_velocity"]["value"]

        assert model.fs.unit.phosphorus_solids_ratio[0].fixed
        assert model.fs.unit.phosphorus_solids_ratio[0].value == data[
            "phosphorus_solids_ratio"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.009864, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.3041363, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["phosphates"]))
        assert (pytest.approx(0.0031360, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(955.68, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["phosphates"]))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (pytest.approx(27.986, rel=1e-5) ==
                value(model.fs.unit.basin_surface_area[0]))
        assert (pytest.approx(19.980, rel=1e-5) ==
                value(model.fs.unit.final_solids_mass[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                      : Value      : Fixed : Bounds
                   Basin Surface Area (ft^2) :     27.986 : False : (None, None)
                          Electricity Demand : 7.0000e-10 : False : (0, None)
                       Electricity Intensity :     0.0000 :  True : (None, None)
    Final mass flow of settled solids (kg/s) :     19.980 : False : (None, None)
             Phosphorus-Solids Ratio (kg/kg) :    0.15000 :  True : (None, None)
                     Settling Velocity (m/s) :  0.0050000 :  True : (None, None)
                 Solute Removal [phosphates] :    0.99900 :  True : (0, None)
                              Water Recovery :    0.98610 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                    Inlet    Treated  Byproduct
    Volumetric Flowrate           0.013000 0.0098640 0.0031360 
    Mass Concentration H2O          769.23    999.70    44.324 
    Mass Concentration phosphates   230.77   0.30414    955.68 
====================================================================================
"""

        assert output in stream.getvalue()


db = Database()
params = db._get_technology("sedimentation")

class TestSedimentationZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tds", "tss"]})

        m.fs.unit = SedimentationZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("sedimentation", subtype=subtype)

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            if j not in data["removal_frac_mass_solute"].keys():
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]


@pytest.mark.parametrize("subtype", [k for k in params.keys()])
@pytest.mark.component
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = SedimentationZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.sedimentation, Block)
    assert isinstance(m.fs.costing.sedimentation.capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.sedimentation.capital_b_parameter,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]

@pytest.mark.unit
def test_phosphorus_capture_no_tss_or_phosphate_in_solute_list_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["foo"]})

    with pytest.raises(KeyError,
                       match="One of the following should be specified in the solute_list: "
                             "tss or phosphates."):
        m.fs.unit = SedimentationZO(default={
            "property_package": m.fs.params,
            "database": db,
            "process_subtype": "phosphorus_capture"})

@pytest.mark.unit
def test_phosphorus_capture_phosphate_tss_in_solute_list_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["tss", "phosphates"]})

    with pytest.raises(KeyError,
                       match="tss and phosphates cannot both be defined in the "
                             "solute_list. Please choose one."):
        m.fs.unit = SedimentationZO(default={
            "property_package": m.fs.params,
            "database": db,
            "process_subtype": "phosphorus_capture"})



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order primary separator model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    Block, ConcreteModel, Constraint, value, Var, assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import PrimarySeparatorZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()

class TestPrimarySeparatorZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["oil_and_grease", "oily_matter", "tss"]})

        m.fs.unit = PrimarySeparatorZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "oil_and_grease"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "oily_matter"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "primary_separator"
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("primary_separator")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]


    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.3e-2, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(76.923, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["oil_and_grease"]))
        assert (pytest.approx(76.923, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["oily_matter"]))
        assert (pytest.approx(76.923, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.010199174, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(9.8047, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["oil_and_grease"]))
        assert (pytest.approx(9.8047, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["oily_matter"]))
        assert (pytest.approx(0.89948, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(2.8008e-3, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(321.334, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["oil_and_grease"]))
        assert (pytest.approx(321.334, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["oily_matter"]))
        assert (pytest.approx(353.76, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                             : Value      : Fixed : Bounds
                 Electricity Demand : 7.0000e-10 : False : (0, None)
              Electricity Intensity :     0.0000 :  True : (None, None)
    Solute Removal [oil_and_grease] :    0.90000 :  True : (0, None)
       Solute Removal [oily_matter] :    0.90000 :  True : (0, None)
               Solute Removal [tss] :    0.99083 :  True : (0, None)
                     Water Recovery :    0.99900 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                        Inlet   Treated  Byproduct
    Volumetric Flowrate               0.013000 0.010199 0.0028008 
    Mass Concentration H2O              769.23   979.49    3.5704 
    Mass Concentration oil_and_grease   76.923   9.8047    321.33 
    Mass Concentration oily_matter      76.923   9.8047    321.33 
    Mass Concentration tss              76.923  0.89948    353.76 
====================================================================================
"""

        assert output in stream.getvalue()

class TestPrimarySeparatorZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["oil_and_grease", "oily_matter", "tss", "foo"]})

        m.fs.unit = PrimarySeparatorZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "oil_and_grease"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "oily_matter"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "primary_separator"
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("primary_separator")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.4e-2, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(71.429, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["oil_and_grease"]))
        assert (pytest.approx(71.429, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["oily_matter"]))
        assert (pytest.approx(71.429, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(71.429, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.011199174, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(8.9292, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["oil_and_grease"]))
        assert (pytest.approx(8.9292, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["oily_matter"]))
        assert (pytest.approx(0.81917, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(89.292, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(2.800826e-3, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(321.334, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["oil_and_grease"]))
        assert (pytest.approx(321.334, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["oily_matter"]))
        assert (pytest.approx(353.76, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(2.8563e-07, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                             : Value      : Fixed : Bounds
                 Electricity Demand : 8.0000e-10 : False : (0, None)
              Electricity Intensity :     0.0000 :  True : (None, None)
               Solute Removal [foo] :     0.0000 :  True : (0, None)
    Solute Removal [oil_and_grease] :    0.90000 :  True : (0, None)
       Solute Removal [oily_matter] :    0.90000 :  True : (0, None)
               Solute Removal [tss] :    0.99083 :  True : (0, None)
                     Water Recovery :    0.99900 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                        Inlet   Treated  Byproduct
    Volumetric Flowrate               0.014000 0.011199  0.0028008
    Mass Concentration H2O              714.29   892.03     3.5704
    Mass Concentration oil_and_grease   71.429   8.9292     321.33
    Mass Concentration oily_matter      71.429   8.9292     321.33
    Mass Concentration tss              71.429  0.81917     353.76
    Mass Concentration foo              71.429   89.292 2.8563e-07
====================================================================================
"""

        assert output in stream.getvalue()

def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = PrimarySeparatorZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.primary_separator, Block)
    assert isinstance(m.fs.costing.primary_separator.capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.primary_separator.capital_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.primary_separator.reference_state, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for general zero-order property package
"""
import pytest
from io import StringIO

from idaes.core import declare_process_block_class, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import get_solver
from pyomo.environ import (ConcreteModel,
                           value)
from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent

from watertap.core import WaterParameterBlock, WaterStateBlock, ZeroOrderBaseData
from watertap.core.zero_order_pt import (
    build_pt, initialize_pt, calculate_scaling_factors_pt, _get_Q_pt)

solver = get_solver()


@declare_process_block_class("DerivedPT")
class DerivedPTData(ZeroOrderBaseData):
    def build(self):
        super().build()

        build_pt(self)


class TestPT:
    @pytest.fixture(scope="module")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        m.fs.unit = DerivedPT(
            default={"property_package": m.fs.water_props})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "A"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "B"].fix(20)
        m.fs.unit.inlet.flow_mass_comp[0, "C"].fix(30)

        return m

    @pytest.mark.unit
    def test_private_attributes(self, model):
        assert model.fs.unit._tech_type is None
        assert model.fs.unit._has_recovery_removal is False
        assert model.fs.unit._fixed_perf_vars == []
        assert model.fs.unit._initialize is initialize_pt
        assert model.fs.unit._scaling is calculate_scaling_factors_pt
        assert model.fs.unit._get_Q is _get_Q_pt
        assert model.fs.unit._stream_table_dict == {
            "Inlet": model.fs.unit.inlet,
            "Outlet": model.fs.unit.outlet}
        assert model.fs.unit._perf_var_dict == {}

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.properties, WaterStateBlock)

        assert isinstance(model.fs.unit.inlet, Port)
        assert isinstance(model.fs.unit.outlet, Port)

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    # Nothing to initialize or solve

    @pytest.mark.component
    def test_solution(self, model):
        for (t, j), v in model.fs.unit.outlet.flow_mass_comp.items():
            assert (pytest.approx(value(
                model.fs.unit.inlet.flow_mass_comp[t, j]), rel=1e-5) ==
                value(v))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()
        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance


------------------------------------------------------------------------------------
    Stream Table
                            Inlet  Outlet
    Volumetric Flowrate    1.0600  1.0600
    Mass Concentration H2O 943.40  943.40
    Mass Concentration A   9.4340  9.4340
    Mass Concentration B   18.868  18.868
    Mass Concentration C   28.302  28.302
====================================================================================
"""

        assert output == stream.getvalue()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import sys
import pytest
from io import StringIO

from pyomo.environ import (ConcreteModel, TransformationFactory,
                           assert_optimal_termination)
from pyomo.network import Arc
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.initialization import propagate_state

# Import components
from watertap.unit_models.mvc.components import Evaporator
from watertap.unit_models.mvc.components import Compressor
# Import property packages
import watertap.property_models.seawater_prop_pack as props_sw
import watertap.property_models.water_prop_pack as props_w

def build(m):
    # Properties
    m.fs.properties_feed = props_sw.SeawaterParameterBlock()
    m.fs.properties_vapor = props_w.WaterParameterBlock()

    # Evaporator
    m.fs.evaporator = Evaporator(default={"property_package_feed": m.fs.properties_feed,
                                          "property_package_vapor": m.fs.properties_vapor,})
    # Compressor
    m.fs.compressor = Compressor(default={"property_package": m.fs.properties_vapor})

    # Connections
    m.fs.s01 = Arc(source=m.fs.evaporator.outlet_vapor, destination=m.fs.compressor.inlet)
    m.fs.s02 = Arc(source=m.fs.compressor.outlet, destination=m.fs.evaporator.inlet_condenser)
    TransformationFactory("network.expand_arcs").apply_to(m)


def scale(m):
    m.fs.properties_feed.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
    m.fs.properties_feed.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'TDS'))
    m.fs.properties_vapor.set_default_scaling('flow_mass_phase_comp', 1, index=('Vap', 'H2O'))
    m.fs.properties_vapor.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
    # Evaporator
    iscale.set_scaling_factor(m.fs.evaporator.area, 1e-3)
    iscale.set_scaling_factor(m.fs.evaporator.U, 1e-3)
    iscale.set_scaling_factor(m.fs.evaporator.delta_temperature_in, 1e-1)
    iscale.set_scaling_factor(m.fs.evaporator.delta_temperature_out, 1e-1)
    iscale.set_scaling_factor(m.fs.evaporator.lmtd, 1e-1)
    # iscale.set_scaling_factor(m.fs.evaporator.heat_transfer, 1e-6)
    # Compressor
    iscale.set_scaling_factor(m.fs.compressor.control_volume.work, 1e-6)

    iscale.calculate_scaling_factors(m)


def specify(m):
    # state variables
    # Feed inlet
    m.fs.evaporator.inlet_feed.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(1)
    m.fs.evaporator.inlet_feed.flow_mass_phase_comp[0, 'Liq', 'TDS'].fix(0.05)
    m.fs.evaporator.inlet_feed.temperature[0].fix(273.15 + 50.52)  # K
    m.fs.evaporator.inlet_feed.pressure[0].fix(1e5)  # Pa

    m.fs.evaporator.outlet_vapor.flow_mass_phase_comp[0, 'Vap', 'H2O'].fix(0.5)
    m.fs.evaporator.U.fix(1e3)  # W/K-m^2
    m.fs.evaporator.area.fix(100)  # m^2

    m.fs.compressor.pressure_ratio.fix(2)
    m.fs.compressor.efficiency.fix(0.8)


def initialize(m):
    m.fs.evaporator.inlet_condenser.flow_mass_phase_comp[0, 'Vap', 'H2O'].fix(0.5)
    m.fs.evaporator.inlet_condenser.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(1e-8)
    m.fs.evaporator.inlet_condenser.temperature[0].fix(400)  # K
    m.fs.evaporator.inlet_condenser.pressure[0].fix(0.5e5)  # Pa

    m.fs.evaporator.initialize(outlvl=idaeslog.INFO_HIGH)

    m.fs.evaporator.inlet_condenser.flow_mass_phase_comp[0, 'Vap', 'H2O'].unfix()
    m.fs.evaporator.inlet_condenser.flow_mass_phase_comp[0, 'Liq', 'H2O'].unfix()
    m.fs.evaporator.inlet_condenser.temperature[0].unfix()  # K
    m.fs.evaporator.inlet_condenser.pressure[0].unfix()  # Pa

    propagate_state(m.fs.s01)
    m.fs.compressor.initialize(outlvl=idaeslog.INFO_HIGH)

@pytest.mark.requires_idaes_solver
@pytest.mark.component
def test_mvc():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    build(m)
    assert_units_consistent(m)
    scale(m)
    specify(m)
    assert (degrees_of_freedom(m) == 0)

    initialize(m)

    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    m.fs.compressor.report()
    m.fs.evaporator.condenser.report()
    m.fs.evaporator.display()
    brine_blk = m.fs.evaporator.feed_side.properties_brine[0]
    # evaporator values
    assert brine_blk.pressure.value == pytest.approx(2.2738e4, rel=1e-3)
    assert m.fs.evaporator.lmtd.value == pytest.approx(12.30, rel=1e-3)
    assert m.fs.evaporator.feed_side.heat_transfer.value == pytest.approx(1.231e6, rel=1e-3)

    # compressor values
    compressed_blk = m.fs.compressor.control_volume.properties_out[0]
    assert m.fs.compressor.control_volume.work[0].value == pytest.approx(5.843e4, rel=1e-3)
    assert compressed_blk.pressure.value == pytest.approx(4.548e4, rel=1e-3)
    assert compressed_blk.temperature.value == pytest.approx(412.98, rel=1e-3)

    # condenser values
    condensed_blk = m.fs.evaporator.condenser.control_volume.properties_out[0]
    assert m.fs.evaporator.condenser.control_volume.heat[0].value == pytest.approx(-1.231e6, rel=1e-3)
    assert condensed_blk.temperature.value == pytest.approx(337.95, rel=1e-3)




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a Ozone reactor unit.
"""

from pyomo.environ import units as pyunits, Var
from pyomo.common.config import ConfigValue, In
from idaes.core.util.exceptions import ConfigurationError
from idaes.core import declare_process_block_class
from watertap.core import build_siso, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("OzoneZO")
class OzoneZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Ozone unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "ozonation"

        build_siso(self)

        if 'toc' not in self.config.property_package.config.solute_list:
            raise ConfigurationError(
                "TOC must be in solute list for Ozonation or Ozone/AOP")

        self.contact_time = Var(self.flowsheet().time,
                                units=pyunits.minute,
                                doc="Ozone contact time")

        self.concentration_time = Var(
            self.flowsheet().time,
            units=(pyunits.mg*pyunits.minute)/pyunits.liter,
            doc="CT value for ozone contactor")

        self.mass_transfer_efficiency = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Ozone mass transfer efficiency")

        self.specific_energy_coeff = Var(
            self.flowsheet().time,
            units=pyunits.kWh/pyunits.lb,
            bounds=(0, None),
            doc="Specific energy consumption for ozone generation")

        self._fixed_perf_vars.append(self.contact_time)
        self._fixed_perf_vars.append(self.concentration_time)
        self._fixed_perf_vars.append(self.mass_transfer_efficiency)
        self._fixed_perf_vars.append(self.specific_energy_coeff)

        self.ozone_flow_mass = Var(self.flowsheet().time,
                                   initialize=1,
                                   bounds=(0, None),
                                   units=pyunits.lb/pyunits.hr,
                                   doc="Mass flow rate of ozone")

        self.ozone_consumption = Var(self.flowsheet().time,
                                     initialize=1,
                                     bounds=(0, None),
                                     units=pyunits.mg/pyunits.liter,
                                     doc="Ozone consumption")

        self.electricity = Var(self.flowsheet().time,
                               initialize=1,
                               bounds=(0, None),
                               units=pyunits.kW,
                               doc="Ozone generation power demand")

        @self.Constraint(self.flowsheet().time,
                         doc="Ozone consumption constraint")
        def ozone_consumption_constraint(b, t):
            return (b.ozone_consumption[t] == (
                (pyunits.convert(b.properties_in[t].conc_mass_comp['toc'],
                                 to_units=pyunits.mg/pyunits.liter) +
                 self.concentration_time[t] / self.contact_time[t])) /
                self.mass_transfer_efficiency[t])

        @self.Constraint(self.flowsheet().time,
                         doc='Ozone mass flow constraint')
        def ozone_flow_mass_constraint(b, t):
            return b.ozone_flow_mass[t] == pyunits.convert(
                b.properties_in[t].flow_vol*b.ozone_consumption[t],
                to_units=pyunits.lb/pyunits.hr)

        @self.Constraint(self.flowsheet().time,
                         doc='Ozone power constraint')
        def electricity_constraint(b, t):
            return b.electricity[t] == (
                b.specific_energy_coeff[t] * b.ozone_flow_mass[t])

        self._perf_var_dict["Ozone Contact Time (min)"] = self.contact_time
        self._perf_var_dict["Ozone CT Value ((mg*min)/L)"] = \
            self.concentration_time
        self._perf_var_dict["Ozone Mass Transfer Efficiency"] = \
            self.mass_transfer_efficiency
        self._perf_var_dict["Ozone Mass Flow (lb/hr)"] = self.ozone_flow_mass
        self._perf_var_dict["Ozone Unit Power Demand (kW)"] = \
            self.electricity



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a cooling tower unit.
"""

from pyomo.environ import Constraint, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("CoolingTowerZO")
class CoolingTowerZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a cooling tower unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "cooling_tower"

        build_sido(self)
        constant_intensity(self)

        self.cycles = Var(units=pyunits.dimensionless,
                          doc="Cycles of concentration")

        self.blowdown = Var(self.flowsheet().time,
                                units=pyunits.m**3/pyunits.hour,
                                doc="Flowrate of blowdown")


        self._fixed_perf_vars.append(self.cycles)

        @self.Constraint(self.flowsheet().time,
                    doc='Blowdown constraint')
        def blowdown_constraint(b, t):
            q_in = pyunits.convert(b.properties_in[t].flow_vol,
                                   to_units=pyunits.m**3/pyunits.hour)
            return b.blowdown[t] == pyunits.convert(q_in*b.cycles, to_units=pyunits.m**3/pyunits.hour)

        self._perf_var_dict["Cycles"] = self.cycles
        self._perf_var_dict["Blowdown flowrate (m^3/hr)"] = self.blowdown



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
from pyomo.environ import (ConcreteModel,
                           Constraint,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           Var)
from pyomo.network import Port
from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)
from watertap.unit_models.crystallizer import Crystallization
import watertap.property_models.cryst_prop_pack as props

from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     unscaled_variables_generator,
                                     unscaled_constraints_generator,
                                     badly_scaled_var_generator)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestCrystallization():
    @pytest.fixture(scope="class")
    def Crystallizer_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = Crystallization(default={
            "property_package": m.fs.properties,
            })

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_NaCl = 0.2126
        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        feed_pressure = 101325
        feed_temperature = 273.15 + 20
        eps = 1e-6
        crystallizer_temperature = 273.15 + 55
        crystallizer_yield = 0.40

        # Fully define feed
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(feed_flow_mass * feed_mass_frac_NaCl)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(feed_flow_mass * feed_mass_frac_H2O)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Sol', 'NaCl'].fix(eps)
        m.fs.unit.inlet.flow_mass_phase_comp[0, 'Vap', 'H2O'].fix(eps)        
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)

        # Define operating conditions
        m.fs.unit.temperature_operating.fix(crystallizer_temperature)
        m.fs.unit.crystallization_yield['NaCl'].fix(crystallizer_yield)

        # Fix growth rate, crystal length and Sounders brown constant to default values
        m.fs.unit.crystal_growth_rate.fix()
        m.fs.unit.souders_brown_constant.fix()
        m.fs.unit.crystal_median_length.fix()

        return m

    @pytest.mark.unit
    def test_config(self, Crystallizer_frame):
        m = Crystallizer_frame
        # check unit config arguments
        assert len(m.fs.unit.config) == 4

        assert not m.fs.unit.config.dynamic
        assert not m.fs.unit.config.has_holdup
        assert m.fs.unit.config.property_package is \
               m.fs.properties

    @pytest.mark.unit
    def test_build(self, Crystallizer_frame):
        m = Crystallizer_frame

        # test ports and variables
        port_lst = ['inlet', 'outlet', 'solids', 'vapor']
        port_vars_lst = ['flow_mass_phase_comp', 'pressure', 'temperature']
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        # First, parameters
        unit_objs_params_lst = ['approach_temperature_heat_exchanger', 'dimensionless_crystal_length']
        for obj_str in unit_objs_params_lst:
            assert hasattr(m.fs.unit, obj_str)
        # Next, variables
        unit_objs_vars_lst = ['crystal_growth_rate', 'crystal_median_length', 'crystallization_yield', 'dens_mass_magma', \
                            'dens_mass_slurry', 'diameter_crystallizer', 'height_crystallizer', 'height_slurry',
                            'magma_circulation_flow_vol', 'pressure_operating', 'product_volumetric_solids_fraction',
                            'relative_supersaturation', 'souders_brown_constant', 't_res', 'temperature_operating',
                            'volume_suspension', 'work_mechanical']
        for obj_str in unit_objs_vars_lst:
            assert hasattr(m.fs.unit, obj_str)
        # Next, expressions
        unit_objs_expr_lst = ['eq_max_allowable_velocity', 'eq_minimum_height_diameter_ratio', 'eq_vapor_space_height']
        for obj_str in unit_objs_expr_lst:
            assert hasattr(m.fs.unit, obj_str)
        # Finally, constraints
        unit_objs_cons_lst = ['eq_T_con1', 'eq_T_con2', 'eq_T_con3', 'eq_crystallizer_height_constraint', 'eq_dens_magma', \
                            'eq_dens_mass_slurry', 'eq_enthalpy_balance', 'eq_mass_balance_constraints', 'eq_minimum_hex_circulation_rate_constraint', 'eq_operating_pressure_constraint',
                            'eq_p_con1', 'eq_p_con2', 'eq_p_con3', 'eq_relative_supersaturation', 'eq_removal_balance',
                            'eq_residence_time', 'eq_slurry_height_constraint', 'eq_solubility_massfrac_equality_constraint', 'eq_suspension_volume', 'eq_vapor_head_diameter_constraint',
                            'eq_vol_fraction_solids'
                            ]
        for obj_str in unit_objs_cons_lst:
            assert hasattr(m.fs.unit, obj_str)


        # Test stateblocks 
        # List olf attributes on all stateblocks
        stateblock_objs_lst = \
            ['flow_mass_phase_comp', 'pressure', 'temperature', 'solubility_mass_phase_comp', 'solubility_mass_frac_phase_comp',
             'mass_frac_phase_comp', 'dens_mass_solvent', 'dens_mass_solute', 'dens_mass_phase', 'cp_phase', 'cp_solvent',
             'flow_vol_phase', 'flow_vol', 'enth_flow', 'enth_mass_solvent', 'dh_crystallization',
             'eq_solubility_mass_phase_comp', 'eq_solubility_mass_frac_phase_comp',
             'eq_mass_frac_phase_comp', 'eq_dens_mass_solvent', 'eq_dens_mass_solute', 'eq_dens_mass_phase', 'eq_cp_solute', 'eq_cp_phase',
             'eq_flow_vol_phase', 'eq_enth_mass_solvent']
        # List of attributes for liquid stateblocks only
        stateblock_objs_liq_lst = ['pressure_sat', 'eq_pressure_sat']
        # Inlet block
        assert hasattr(m.fs.unit, 'properties_in')
        blk = getattr(m.fs.unit, 'properties_in')
        for var_str in stateblock_objs_lst:
            assert hasattr(blk[0], var_str)
        for var_str in stateblock_objs_liq_lst:
            assert hasattr(blk[0], var_str)

        # Liquid outlet block
        assert hasattr(m.fs.unit, 'properties_out')
        blk = getattr(m.fs.unit, 'properties_out')
        for var_str in stateblock_objs_lst:
            assert hasattr(blk[0], var_str)
        for var_str in stateblock_objs_liq_lst:
            assert hasattr(blk[0], var_str)

        # Vapor outlet block
        assert hasattr(m.fs.unit, 'properties_vapor')
        blk = getattr(m.fs.unit, 'properties_vapor')
        for var_str in stateblock_objs_lst:
            assert hasattr(blk[0], var_str)

        # Liquid outlet block
        assert hasattr(m.fs.unit, 'properties_solids')
        blk = getattr(m.fs.unit, 'properties_solids')
        for var_str in stateblock_objs_lst:
            assert hasattr(blk[0], var_str)

        # test statistics
        assert number_variables(m) == 238
        assert number_total_constraints(m) == 124
        assert number_unused_variables(m) == 4  

    @pytest.mark.unit
    def test_dof(self, Crystallizer_frame):
        m = Crystallizer_frame
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, Crystallizer_frame):
        m = Crystallizer_frame

        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Liq', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Liq', 'NaCl'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Vap', 'H2O'))
        m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Sol', 'NaCl'))
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

        for _ in badly_scaled_var_generator(m):
            assert False    

    @pytest.mark.component
    def test_initialize(self, Crystallizer_frame):
        initialization_tester(Crystallizer_frame)

    # @pytest.mark.component
    # def test_var_scaling(self, Crystallizer_frame):
    #     m = Crystallizer_frame
    #     badly_scaled_var_lst = list(badly_scaled_var_generator(m))
    #     assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, Crystallizer_frame):
        m = Crystallizer_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert results.solver.termination_condition == \
               TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_conservation(self, Crystallizer_frame):
        m = Crystallizer_frame
        b = m.fs.unit
        comp_lst = ['NaCl', 'H2O']
        phase_lst = ['Sol', 'Liq', 'Vap']
        phase_comp_list = [(p, j) for j in comp_lst for p in phase_lst if (p, j) in b.properties_in[0].phase_component_set]

        flow_mass_in = sum(b.properties_in[0].flow_mass_phase_comp[p, j] for p in phase_lst 
            for j in comp_lst 
            if (p, j) in phase_comp_list)
        flow_mass_out = sum(b.properties_out[0].flow_mass_phase_comp[p, j] for p in phase_lst 
            for j in comp_lst 
            if (p, j) in phase_comp_list)
        flow_mass_solids = sum(b.properties_solids[0].flow_mass_phase_comp[p, j] for p in phase_lst 
            for j in comp_lst 
            if (p, j) in phase_comp_list)
        flow_mass_vapor = sum(b.properties_vapor[0].flow_mass_phase_comp[p, j] for p in phase_lst 
            for j in comp_lst 
            if (p, j) in phase_comp_list)

        assert (abs(value(flow_mass_in - flow_mass_out - flow_mass_solids - flow_mass_vapor
                          )) <= 1e-6)

        assert (abs(value(
            flow_mass_in * b.properties_in[0].enth_mass_phase['Liq']
            - flow_mass_out * b.properties_out[0].enth_mass_phase['Liq']
            - flow_mass_vapor * b.properties_vapor[0].enth_mass_solvent['Vap']
            - flow_mass_solids * b.properties_solids[0].enth_mass_solute['Sol']
            - flow_mass_solids * b.properties_solids[0].dh_crystallization['NaCl']
            + b.work_mechanical[0]
        )) <= 1e-2)

    @pytest.mark.component
    def test_solution(self, Crystallizer_frame):
        m = Crystallizer_frame
        b = m.fs.unit
        # Check solid mass in solids stream
        assert (pytest.approx(value(b.crystallization_yield['NaCl'] * b.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl']), rel=1e-3) ==
                value(b.solids.flow_mass_phase_comp[0, 'Sol', 'NaCl'])
                )    
        # Check solid mass in liquid stream   
        assert (pytest.approx(value((1-b.crystallization_yield['NaCl']) * b.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl']), rel=1e-3) ==
                value(b.outlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'])
                )    
        # Check outlet liquid stream composition which is set by solubility
        assert (pytest.approx(0.2695, rel=1e-3) ==
                value(b.properties_out[0].mass_frac_phase_comp['Liq', 'NaCl'])
                )    
        # Check liquid stream solvent flow
        assert (pytest.approx(0.12756 * ((1/0.2695) - 1), rel=1e-3) ==
                value(b.outlet.flow_mass_phase_comp[0, 'Liq', 'H2O'])
                ) 
        # Check saturation pressure 
        assert (pytest.approx(11992, rel=1e-3) ==
                value(b.pressure_operating)
                ) 
        # Check heat requirement
        assert (pytest.approx(1127.2, rel=1e-3) ==
                value(b.work_mechanical[0])
                ) 
        # Check crystallizer diameter
        assert (pytest.approx(1.205, rel=1e-3) ==
                value(b.diameter_crystallizer)
                ) 
        # Minimum active volume
        assert (pytest.approx(1.619, rel=1e-3) ==
                value(b.volume_suspension)
                ) 
        # Residence time
        assert (pytest.approx(1.0228, rel=1e-3) ==
                value(b.t_res)
                ) 

    @pytest.mark.component
    def test_solution2(self, Crystallizer_frame):
        m = Crystallizer_frame
        b = m.fs.unit
        b.crystal_growth_rate.fix(5e-8)
        b.souders_brown_constant.fix(0.0244)
        b.crystal_median_length.fix(0.4e-3)
        results = solver.solve(m)
        
        # Test that report function works
        b.report()

        # Check for optimal solution
        assert results.solver.termination_condition == \
               TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        # Residence time
        assert (pytest.approx(value(b.crystal_median_length / (3.67 * 3600 * b.crystal_growth_rate)) , rel=1e-3) ==
                value(b.t_res)
                ) 
        # Check crystallizer diameter
        assert (pytest.approx(1.5427, rel=1e-3) ==
                value(b.diameter_crystallizer)
                ) 
        # Minimum active volume
        assert (pytest.approx(0.959, rel=1e-3) ==
                value(b.volume_suspension)
                ) 



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order anaerobic MBR-MEC model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    ConcreteModel, Constraint, value, Var, assert_optimal_termination, units as pyunits)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester

from watertap.unit_models.zero_order import AnaerobicMBRMECZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock

solver = get_solver()

class TestAnaerobicMBRMECZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["cod",
                                     "nonbiodegradable_cod",
                                     "ammonium_as_nitrogen", "phosphate_as_phosphorous"]})

        m.fs.unit = AnaerobicMBRMECZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(0.043642594)
        m.fs.unit.inlet.flow_mass_comp[0, "cod"].fix(1.00625e-4)
        m.fs.unit.inlet.flow_mass_comp[0, "nonbiodegradable_cod"].fix(1e-20)
        m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(4.59375E-06)
        m.fs.unit.inlet.flow_mass_comp[0, "phosphate_as_phosphorous"].fix(2.1875E-06)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("anaerobic_mbr_mec")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data['removal_frac_mass_solute'].keys():
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(3780/3600/24/1000, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(2.3, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["cod"]))
        assert (pytest.approx(0, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(0.105, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["ammonium_as_nitrogen"]))
        assert (pytest.approx(0.05, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["phosphate_as_phosphorous"]))

        assert (pytest.approx(1500/3600/24/1000, rel=1e-2) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(2.8958, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["cod"]))
        assert (pytest.approx(4.60445e-05, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonbiodegradable_cod"]))

        assert (pytest.approx(2280/3600/24/1000, rel=1e-2) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(3.0331e-05, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["cod"]))
        assert (pytest.approx(1.90757, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(4.347, rel=1e-3) ==
                value(pyunits.convert(
                    model.fs.unit.properties_byproduct[0].flow_mass_comp["nonbiodegradable_cod"],
                    to_units=pyunits.kg/pyunits.day)))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] +
                sum(model.fs.unit.generation_rxn_comp[0, r, j]
                    for r in model.fs.unit.reaction_set) -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                           : Value      : Fixed : Bounds
                               Electricity Demand : 8.0000e-10 : False : (0, None)
                            Electricity Intensity :     0.0000 :  True : (None, None)
    Reaction Extent [cod_to_nonbiodegradable_cod] : 5.0313e-05 : False : (None, None)
            Solute Removal [ammonium_as_nitrogen] :     0.0000 :  True : (0, None)
                             Solute Removal [cod] :     0.0000 :  True : (0, None)
            Solute Removal [nonbiodegradable_cod] :     1.0000 :  True : (0, None)
        Solute Removal [phosphate_as_phosphorous] :     0.0000 :  True : (0, None)
                                   Water Recovery :    0.39680 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                                   Inlet     Treated   Byproduct
    Volumetric Flowrate                         4.3750e-05 1.7374e-05 2.6376e-05
    Mass Concentration H2O                          997.55     996.71     998.09
    Mass Concentration cod                          2.3000     2.8958 3.0331e-05
    Mass Concentration nonbiodegradable_cod     2.2857e-16 4.6045e-05     1.9076
    Mass Concentration ammonium_as_nitrogen        0.10500    0.26444 3.0331e-05
    Mass Concentration phosphate_as_phosphorous   0.050000    0.12595 3.0331e-05
====================================================================================
"""

        assert output in stream.getvalue()


db = Database()
params = db._get_technology("anaerobic_mbr_mec")


class Test_AnMBRMEC_ZO_subtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["cod", "nonbiodegradable_cod"]})

        m.fs.unit = AnaerobicMBRMECZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("anaerobic_mbr_mec", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]


@pytest.mark.unit
def test_ffCOD_not_in_solute_list():
    model = ConcreteModel()
    model.db = Database()

    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.params = WaterParameterBlock(default={"solute_list": ["cod"]})
    with pytest.raises(ValueError,
                       match="nonbiodegradable_cod must be included in the solute list since"
                             " this unit model converts cod to nonbiodegradable_cod."):
        model.fs.unit = AnaerobicMBRMECZO(default={"property_package": model.fs.params,
                                                   "database": model.db})


@pytest.mark.unit
def test_COD_not_in_solute_list():
    model = ConcreteModel()
    model.db = Database()

    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.params = WaterParameterBlock(default={"solute_list": ["nonbiodegradable_cod"]})
    with pytest.raises(ValueError,
                       match="fs.unit - key_reactant cod for reaction cod_to_nonbiodegradable_cod " 
                             "is not in the component list used by the assigned property package."):
        model.fs.unit = AnaerobicMBRMECZO(default={"property_package": model.fs.params,
                                                   "database": model.db})


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order coagulation/flocculation model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    Block, ConcreteModel, Constraint, value, Var, assert_optimal_termination, units as pyunits)
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import CoagulationFlocculationZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestCoagFlocZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tss"]})

        m.fs.unit = CoagulationFlocculationZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        vars = {'alum_dose',
                'polymer_dose',
                'anion_to_cation_polymer_ratio',
                'chemical_flow_mass',
                'chemical_flow_mass',
                'rapid_mix_retention_time',
                'floc_retention_time',
                'rapid_mix_basin_vol',
                'floc_basin_vol',
                'num_rapid_mixers',
                'num_floc_mixers',
                'num_rapid_mix_processes',
                'num_floc_processes',
                'num_coag_processes',
                'num_floc_injection_processes',
                'velocity_gradient_rapid_mix',
                'velocity_gradient_floc',
                'power_rapid_mix',
                'power_floc',
                'anionic_polymer_dose',
                'cationic_polymer_dose',
                'electricity'}
        cons = {'rapid_mix_basin_vol_constraint',
                'floc_basin_vol_constraint',
                'chemical_flow_constraint',
                'chemical_flow_constraint',
                'rule_power_rapid_mix',
                'rule_power_floc',
                'anionic_polymer_dose_constraint',
                'cationic_polymer_dose_constraint',
                'electricity_constraint'}
        for var in vars:
            assert isinstance(getattr(model.fs.unit, var), Var)
        for con in cons:
            assert isinstance(getattr(model.fs.unit, con), Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("coag_and_floc")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.alum_dose[0].fixed
        assert value(model.fs.unit.alum_dose[0]) == data['alum_dose']['value']

        assert model.fs.unit.polymer_dose[0].fixed
        assert value(model.fs.unit.polymer_dose[0]) == data['polymer_dose']['value']

        assert model.fs.unit.anion_to_cation_polymer_ratio[0].fixed
        assert value(model.fs.unit.anion_to_cation_polymer_ratio[0]) == \
               data['anion_to_cation_polymer_ratio']['value']

        assert model.fs.unit.rapid_mix_retention_time[0].fixed
        assert value(model.fs.unit.rapid_mix_retention_time[0]) == \
               data['rapid_mix_retention_time']['value']


        assert model.fs.unit.floc_retention_time[0].fixed
        assert value(model.fs.unit.floc_retention_time[0]) == \
               data['floc_retention_time']['value']

        assert model.fs.unit.num_rapid_mixers.fixed
        assert value(model.fs.unit.num_rapid_mixers) == \
               data['num_rapid_mixers']['value']

        assert model.fs.unit.num_floc_mixers.fixed
        assert value(model.fs.unit.num_floc_mixers) == \
               data['num_floc_mixers']['value']

        assert model.fs.unit.num_rapid_mix_processes.fixed
        assert value(model.fs.unit.num_rapid_mix_processes) == \
               data['num_rapid_mix_processes']['value']

        assert model.fs.unit.num_floc_processes.fixed
        assert value(model.fs.unit.num_floc_processes) == \
               data['num_floc_processes']['value']

        assert model.fs.unit.num_coag_processes.fixed
        assert value(model.fs.unit.num_coag_processes) == \
               data['num_coag_processes']['value']

        assert model.fs.unit.num_floc_injection_processes.fixed
        assert value(model.fs.unit.num_floc_injection_processes) == \
               data['num_floc_injection_processes']['value']

        assert model.fs.unit.velocity_gradient_rapid_mix[0].fixed
        assert value(model.fs.unit.velocity_gradient_rapid_mix[0]) == \
               data['velocity_gradient_rapid_mix']['value']

        assert model.fs.unit.velocity_gradient_floc[0].fixed
        assert value(model.fs.unit.velocity_gradient_floc[0]) == \
               data['velocity_gradient_floc']['value']

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)
        assert_units_equivalent(model.fs.unit.rapid_mix_retention_time, pyunits.s)
        assert_units_equivalent(model.fs.unit.floc_retention_time, pyunits.min)
        assert_units_equivalent(model.fs.unit.rapid_mix_basin_vol, pyunits.m**3)
        assert_units_equivalent(model.fs.unit.floc_basin_vol, pyunits.m**3)


    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        for t, j in model.fs.unit.inlet.flow_mass_comp:
            assert (pytest.approx(value(
                model.fs.unit.inlet.flow_mass_comp[t, j]), rel=1e-5) ==
                    value(model.fs.unit.outlet.flow_mass_comp[t, j]))

        assert (pytest.approx(0.057915, rel=1e-5) ==
                value(model.fs.unit.power_rapid_mix[0]))

        assert (pytest.approx(0.179712, rel=1e-5) ==
                value(model.fs.unit.power_floc[0]))

        assert (pytest.approx(0.237627, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

        assert (pytest.approx(0.071500, rel=1e-5) ==
                value(model.fs.unit.rapid_mix_basin_vol))

        assert (pytest.approx(9.360, rel=1e-5) ==
                value(model.fs.unit.floc_basin_vol))

        assert (pytest.approx(900, rel=1e-5) ==
                value(model.fs.unit.velocity_gradient_rapid_mix[0]))

        assert (pytest.approx(80, rel=1e-5) ==
                value(model.fs.unit.velocity_gradient_floc[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                               : Value      : Fixed : Bounds
                   Alum Dosage (mg/L) :     10.000 :  True : (0, None)
                     Alum Flow (kg/s) : 0.00013000 : False : (0, None)
              Floc Basin Volume (m^3) :     9.3600 : False : (None, None)
                      Floc Power (kW) :    0.17971 : False : (None, None)
            Floc Retention Time (min) :     12.000 :  True : (None, None)
         Floc Velocity Gradient (1/s) :     80.000 :  True : (None, None)
                Polymer Dosage (mg/L) :    0.10000 :  True : (0, None)
                  Polymer Flow (kg/s) : 1.3000e-06 : False : (0, None)
         Rapid Mix Basin Volume (m^3) :   0.071500 : False : (None, None)
                 Rapid Mix Power (kW) :   0.057915 : False : (None, None)
         Rapid Mix Retention Time (s) :     5.5000 :  True : (None, None)
    Rapid Mix Velocity Gradient (1/s) :     900.00 :  True : (None, None)
         Total Power Consumption (kW) :    0.23763 : False : (None, None)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet   Outlet 
    Volumetric Flowrate    0.013000 0.013000
    Mass Concentration H2O   769.23   769.23
    Mass Concentration tss   230.77   230.77
====================================================================================
"""

        assert output in stream.getvalue()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = CoagulationFlocculationZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.coag_and_floc, Block)
    assert isinstance(m.fs.costing.coag_and_floc.capital_mix_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.coag_and_floc.capital_mix_b_parameter,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for general zero-order property package
"""
import pytest
from io import StringIO

from idaes.core import declare_process_block_class, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import get_solver
from pyomo.environ import (check_optimal_termination,
                           ConcreteModel,
                           Constraint,
                           Param,
                           units as pyunits,
                           value,
                           Var)
from pyomo.util.check_units import assert_units_consistent

from watertap.core import (constant_intensity, pump_electricity,
                           WaterParameterBlock, ZeroOrderBaseData)

solver = get_solver()


def get_Q(self, t):
    # Dummy method to return a flow rate for electricity intensity
    return 42*pyunits.m**3/pyunits.hr


@declare_process_block_class("DerivedZO")
class DerivedZOData(ZeroOrderBaseData):
    def build(self):
        super().build()

        self._get_Q = get_Q


class TestConstantIntensity:
    @pytest.fixture(scope="module")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        m.fs.unit = DerivedZO(
            default={"property_package": m.fs.water_props})

        constant_intensity(m.fs.unit)

        m.fs.unit.energy_electric_flow_vol_inlet.fix(10)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.electricity, Var)
        assert model.fs.unit.electricity.is_indexed()
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert not model.fs.unit.energy_electric_flow_vol_inlet.is_indexed()

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert model.fs.unit.electricity_consumption.is_indexed()

    @pytest.mark.unit
    def test_private_attributes(self, model):
        assert model.fs.unit._tech_type is None
        assert model.fs.unit._has_recovery_removal is False
        assert model.fs.unit._fixed_perf_vars == [
            model.fs.unit.energy_electric_flow_vol_inlet]
        assert model.fs.unit._initialize is None
        assert model.fs.unit._scaling is None
        assert model.fs.unit._get_Q is get_Q
        assert model.fs.unit._perf_var_dict == {
            "Electricity Demand": model.fs.unit.electricity,
            "Electricity Intensity":
                model.fs.unit.energy_electric_flow_vol_inlet}

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(42*10, rel=1e-5) == value(
            model.fs.unit.electricity[0])

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                   : Value  : Fixed : Bounds
       Electricity Demand : 420.00 : False : (0, None)
    Electricity Intensity : 10.000 :  True : (None, None)

------------------------------------------------------------------------------------
    Stream Table
    Empty DataFrame
    Columns: []
    Index: []
====================================================================================
"""

        assert output == stream.getvalue()


class TestPumpElectricity:
    @pytest.fixture(scope="module")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        m.fs.unit = DerivedZO(
            default={"property_package": m.fs.water_props})

        m.fs.unit.flow = Param(m.fs.time,
                               initialize=1,
                               units=pyunits.gallon/pyunits.minute)

        pump_electricity(m.fs.unit, m.fs.unit.flow)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.electricity, Var)
        assert model.fs.unit.electricity.is_indexed()

        assert isinstance(model.fs.unit.lift_height, Param)
        assert value(model.fs.unit.lift_height) == 100
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert value(model.fs.unit.eta_pump) == 0.9
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert value(model.fs.unit.eta_motor) == 0.9

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert model.fs.unit.electricity_consumption.is_indexed()

    @pytest.mark.unit
    def test_private_attributes(self, model):
        assert model.fs.unit._tech_type is None
        assert model.fs.unit._has_recovery_removal is False
        assert model.fs.unit._initialize is None
        assert model.fs.unit._scaling is None
        assert model.fs.unit._get_Q is get_Q
        assert model.fs.unit._perf_var_dict == {
            "Electricity Demand": model.fs.unit.electricity}

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(2.32479e-2, rel=1e-5) == value(
            model.fs.unit.electricity[0])

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                : Value    : Fixed : Bounds
    Electricity Demand : 0.023248 : False : (0, None)

------------------------------------------------------------------------------------
    Stream Table
    Empty DataFrame
    Columns: []
    Index: []
====================================================================================
"""

        assert output == stream.getvalue()